// Copyright 2024 XMOS LIMITED.
// This Software is subject to the terms of the XMOS Public Licence: Version 1.

#include <print.h>

#include "adat_rx.h"

#include "adat_wrapper.h"
#include "utils.h"
#include "app_config.h"
#ifndef ADAT_TX_USE_SHARED_BUFF
#error Designed for ADAT tx shared buffer mode ONLY
#endif
#include "adat_tx.h"

// Global as will be read by utils.xc
uint32_t measured_adat_rate = 0;
unsafe{uint32_t * unsafe adat_rx_sr_ptr = &measured_adat_rate;}

#pragma unsafe arrays // Remove array checks for speed
void adat_rx_demux(streaming chanend c_adat_rx, chanend c_digital_rx_to_asrc, chanend c_smux_change_adat_rx)
{

    unsigned word = 0;                          // Used for first word Rx in select
    adat_state_t adat_state[2] = {{{0}}};
    uint8_t adat_state_wr_idx = 0;              // Double buffer index used for writing rx'd samples
    uint32_t smux_setting = SMUX_NONE;
    unsigned rx_smux_iv_flag = 0;               // Received from GPIO to indicate how to interpret SMUX user bit. If 1, SMUX-IV else SMUX-II

    // Sample rate measurement vars
    timer tmr;
    int32_t rate_measurement_trigger = 0;
    const int32_t rate_measurement_period = XS1_TIMER_HZ / ADAT_RATE_MEASUREMENT_HZ;
    uint32_t rate_sample_count = 0;

    tmr :> rate_measurement_trigger;

    while(1)
    {
        select
        {
            case c_adat_rx :> word:
                // grab timestamp first for minimum jitter
                int32_t time_stamp;
                tmr :> time_stamp;
                // The "other" buffer for reading from
                uint8_t adat_state_rd_idx = adat_state_wr_idx ^ 1;
                // Decode ADAT Rx output
                if(word & 0x01){
                    adat_state[adat_state_wr_idx].sample_number = 0;
                    adat_state[adat_state_wr_idx].user_bits = word >> 4;
                } else {
                    // Store sample in linear array. We expect 8 of these always.
                    adat_state[adat_state_wr_idx].samples[adat_state[adat_state_wr_idx].sample_number] = word;
                    adat_state[adat_state_wr_idx].sample_number++;
                    adat_state[adat_state_wr_idx].rx_time_last_sample = time_stamp;

                    switch(smux_setting){
                        // No SMUX - 12345678
                        case SMUX_NONE:
                            if(adat_state[adat_state_wr_idx].sample_number >= 8){
                                if(measured_adat_rate != 0){
                                    outuint(c_digital_rx_to_asrc, measured_adat_rate);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].rx_time_last_sample);
                                    outuint(c_digital_rx_to_asrc, 8);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[0]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[1]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[2]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[3]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[4]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[5]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[6]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[7]);

                                    adat_state_wr_idx = adat_state_rd_idx;
                                }
                                rate_sample_count++;
                            }
                        break;

                        // SMUX-II 11223344
                        case SMUX_II:
                            if(adat_state[adat_state_wr_idx].sample_number == 4){
                                if(measured_adat_rate != 0){
                                    outuint(c_digital_rx_to_asrc, measured_adat_rate);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].rx_time_last_sample);
                                    outuint(c_digital_rx_to_asrc, 4);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[0]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[2]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[4]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[6]);
                                }
                                rate_sample_count++;
                            }
                            else if(adat_state[adat_state_wr_idx].sample_number >= 8){
                                if(measured_adat_rate != 0){
                                    outuint(c_digital_rx_to_asrc, measured_adat_rate);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].rx_time_last_sample);
                                    outuint(c_digital_rx_to_asrc, 4);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[1]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[3]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[5]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[7]);

                                    adat_state_wr_idx = adat_state_rd_idx;
                                }
                                rate_sample_count++;
                            }
                        break;

                        // SMUX-IV 11112222
                        case SMUX_IV:
                            if(adat_state[adat_state_wr_idx].sample_number == 2){
                                if(measured_adat_rate != 0){
                                    outuint(c_digital_rx_to_asrc, measured_adat_rate);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].rx_time_last_sample);
                                    outuint(c_digital_rx_to_asrc, 2);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[0]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[4]);
                                }
                                rate_sample_count++;
                            }
                            else if(adat_state[adat_state_wr_idx].sample_number == 4){
                                if(measured_adat_rate != 0){
                                    outuint(c_digital_rx_to_asrc, measured_adat_rate);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].rx_time_last_sample);
                                    outuint(c_digital_rx_to_asrc, 2);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[1]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[5]);
                                }
                                rate_sample_count++;
                            }
                            else if(adat_state[adat_state_wr_idx].sample_number == 6){
                                if(measured_adat_rate != 0){
                                    outuint(c_digital_rx_to_asrc, measured_adat_rate);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].rx_time_last_sample);
                                    outuint(c_digital_rx_to_asrc, 2);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[2]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[6]);
                                }
                                rate_sample_count++;
                            }
                            else if(adat_state[adat_state_wr_idx].sample_number >= 8){
                                if(measured_adat_rate != 0){
                                    outuint(c_digital_rx_to_asrc, measured_adat_rate);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].rx_time_last_sample);
                                    outuint(c_digital_rx_to_asrc, 2);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[3]);
                                    outuint(c_digital_rx_to_asrc, adat_state[adat_state_rd_idx].samples[7]);

                                    adat_state_wr_idx = adat_state_rd_idx;
                                }
                                rate_sample_count++;
                            }
                        break;
                    } // Select
                }
            uint32_t current_time;
            tmr :> current_time;
            if((int32_t)(current_time - rate_measurement_trigger) > rate_measurement_period)
            {
                rate_measurement_trigger = current_time;
                // Measure ADAT
                uint32_t samples_per_second = rate_sample_count * ADAT_RATE_MEASUREMENT_HZ;
                measured_adat_rate = get_normal_sample_rate(samples_per_second);
                rate_sample_count = 0;

                // Auto detect if muxed or not and combine with SMUX-IV flag
                if(adat_state[adat_state_wr_idx].user_bits & 0x40){
                    if(rx_smux_iv_flag){
                        smux_setting = SMUX_IV;
                    } else {
                        smux_setting = SMUX_II;
                    }
                } else {
                    smux_setting = SMUX_NONE;
                }
            }

            break;

            case c_smux_change_adat_rx :> word:
                if(word == IO_ADAT_RX){
                    c_smux_change_adat_rx :> rx_smux_iv_flag;
                    // Immediately invalidate rate so we don't send samples at the wrong rate
                    measured_adat_rate = 0;
                    // Ensure we have enough time to properly establish the new rate
                    rate_measurement_trigger += rate_measurement_period;
                }
            break;
        } // select
    } // while
}


void adat_rx_task(streaming chanend c_adat_rx, buffered in port:32 p_adat_in) {
    while(1) {
        adatReceiver48000(p_adat_in, c_adat_rx);
        adatReceiver44100(p_adat_in, c_adat_rx);
    }
}


void adat_tx_hw_setup(chanend c_adat_tx, clock mck_blk, in port p_mclk, buffered out port:32 p_adat_out){
    set_clock_src(mck_blk, p_mclk);
    set_clock_fall_delay(mck_blk, 7);

    set_port_clock(p_adat_out, mck_blk);
    start_clock(mck_blk);
}


void adat_tx_task(chanend c_adat_tx, buffered out port:32 p_adat_out, clock mck_blk, in port p_mclk){
    set_clock_src(mck_blk, p_mclk);
    set_clock_fall_delay(mck_blk, 7);

    set_port_clock(p_adat_out, mck_blk);
    start_clock(mck_blk);

    while(1){
        adat_tx_port(c_adat_tx, p_adat_out);
    }
}


int adat_tx_startup(chanend c_adat_tx, unsigned sample_rate, int32_t *adat_tx_samples){
    // Setup ADAT Tx
    // ADAT parameters ...
    //
    // adat_oversampling =  256 for MCLK = 12M288 or 11M2896
    //                   =  512 for MCLK = 24M576 or 22M5792
    //                   = 1024 for MCLK = 49M152 or 45M1584
    //
    // adatSmuxMode   = 1 for FS =  44K1 or  48K0
    //                = 2 for FS =  88K2 or  96K0
    //                = 4 for FS = 176K4 or 192K0

    int adat_tx_smux = 1;
    if(sample_rate == 0){
        adat_tx_smux = 4; // Set to lowest backpressure setting if invalid sample rate
    }
    else if(sample_rate % 48000 == 0){
        adat_tx_smux = sample_rate / 48000;
    } else {
        adat_tx_smux = sample_rate / 44100;
    }

    outuint(c_adat_tx, ADAT_MULTIPLIER);
    outuint(c_adat_tx, adat_tx_smux);
    // Cast to unsafe pointer then to unsigned so we can pass over channel
    unsafe{
        volatile unsigned * unsafe sample_ptr = (volatile unsigned * unsafe)adat_tx_samples;
        outuint(c_adat_tx, (unsigned) sample_ptr);
    }
    adat_smux_counter = 0;

    return adat_tx_smux;
}


void adat_tx_shutdown(chanend c_adat_tx){
    // Receive outstanding handshake from ADAT core
    inuint(c_adat_tx);
    // Notify ADAT Tx thread of impending new freq...
    outct(c_adat_tx, XS1_CT_END);
}

