// Copyright 2024 XMOS LIMITED.
// This Software is subject to the terms of the XMOS Public Licence: Version 1.

#include "utils.h"
#include "app_config.h"

#if (CONFIG == SPDIF)

// Called from a different tile (ASRC tile) hence channel usage. It is the callback fired by ASRC task on reception of the first channel token.
#pragma unsafe arrays
unsigned spdif_receive_asrc_input_samples(chanend c_digital_rx_to_asrc, asrc_in_out_t &asrc_io, unsigned &new_input_rate){
    // Grab timestamp straight away
    timer tmr;
    int32_t input_timestamp;
    tmr :> input_timestamp;

    const unsigned n_spdif_chans = 2;
    static unsigned asrc_in_counter = 0;    // Used for building SRC_N_IN_SAMPLES worth of samples
    static unsigned input_rate = 0;         // Init to invalid

    unsigned input_dbl_buff_write_idx = asrc_io.input_write_idx;

    asrc_io.input_timestamp[input_dbl_buff_write_idx] = input_timestamp;

    // Check for SR change from SPDIF
    if(testct(c_digital_rx_to_asrc)){
        inct(c_digital_rx_to_asrc); // Discard control token
        input_rate = inuint(c_digital_rx_to_asrc);
        return -1; // No new frame
    }

    // Get SPDIF sample and L/R from channel
    uint32_t frame = inuint(c_digital_rx_to_asrc);
    int spdif_ch = (frame & SPDIF_RX_PREAMBLE_MASK) == SPDIF_FRAME_Y ? 1 : 0;
    int32_t sample = SPDIF_RX_EXTRACT_SAMPLE(frame);

    // Pack nicely into LRLRLRLR
    unsigned buff_idx = spdif_ch + n_spdif_chans * asrc_in_counter;
    asrc_io.input_samples[input_dbl_buff_write_idx][buff_idx] = sample;

    // Set rate
    new_input_rate = input_rate;
    asrc_io.input_channel_count = n_spdif_chans;

    // Move asrc_in_counter on if we have two channels
    if(spdif_ch == n_spdif_chans - 1){
        if(++asrc_in_counter == SRC_N_IN_SAMPLES){
            asrc_in_counter = 0;
            return 0; // New frame
        }
    }

    return -1; // No new frame
}

// Functions used by spdif_rx_custom
int check_clock_div(buffered in port:32 p);
int spdif_rx_decode(streaming chanend c, buffered in port:32 p, unsigned sample_rate);

// Global written to by spdif_rx_custom and read by LED routine in utils
unsafe{
    unsigned spdif_rx_sr = 0;
    unsigned * unsafe spdif_rx_sr_ptr = &spdif_rx_sr;
}

// This is a specialised outer loop for spdif rx which sends the detected sample
// rate to the consumer (ASRC input) and removes support for 32kHz in the SR
// ranges (ASRC doesn't directly support 32kHz input)
void spdif_rx_custom(streaming chanend c, in port p, clock clk, unsigned sample_rate)
{
    int exit = 0;

    in port * movable pp = &p;
    in buffered port:32 * movable p_buf = reconfigure_port(move(pp), in buffered port:32);

    // Configure spdif rx port to be clocked from spdif_rx clock defined below.
    configure_in_port(*p_buf, clk);

    while(1)
    {
        // Determine 100MHz clock divider
        unsigned clock_div = 96001/sample_rate;

        // Stop clock so we can reconfigure it
        stop_clock(clk);

        // Set the desired clock div
        configure_clock_ref(clk, clock_div);

        // Send the control token to indiciate new rate, followed by rate.
        soutct(c, XS1_CT_ACK);
        c <: sample_rate;
        unsafe{ *spdif_rx_sr_ptr = sample_rate;}

        // Start the clock block running. Port timer will be reset here.
        start_clock(clk);

        // Check our clock div value is correct
        if (check_clock_div(*p_buf) == 0)
           exit = spdif_rx_decode(c, *p_buf, sample_rate);

        if(exit)
            break;

        // Get next sample rate from current sample rate.
        switch(sample_rate)
        {
            // case 32000 has been removed
            case 44100:  sample_rate = 48000;  break;
            case 48000:  sample_rate = 88200;  break;
            case 88200:  sample_rate = 96000;  break;
            case 96000:  sample_rate = 176400; break;
            case 176400: sample_rate = 192000; break;
            case 192000: sample_rate = 44100;  break;
            default:     sample_rate = 48000;  break;
        }
    }

    // Set pointers and ownership back to original state if SpdifReceive() exits
    pp = reconfigure_port(move(p_buf), in port);
}

void spdif_rx_task(chanend c_spdif_rx, in port p_spdif_in, clock spdif_rx_clk) {
    unsafe{
        // We need to cast the chanend used by either SPDIF or ADAT to a streaming chanend expected by spdif_rx
        unsafe streaming chanend c_spdif_tx_streaming = c_spdif_rx;
        while(1) {
            spdif_rx_custom((streaming chanend)c_spdif_tx_streaming, p_spdif_in, spdif_rx_clk, 96000);
        }
    }
}

void spdif_tx_task(chanend c_spdif_tx, out buffered port:32 p_spdif_tx, clock clk_spdif_tx, in port p_mclk_in){
    spdif_tx_port_config(p_spdif_tx, clk_spdif_tx, p_mclk_in, 0);
    start_clock(clk_spdif_tx);
    while(1) {
        spdif_tx(p_spdif_tx, c_spdif_tx);
    }
}

void spdif_tx_startup(chanend c_spdif_tx, unsigned sample_rate, unsigned master_clock_frequency){
    if(sample_rate == 0){
        // If invalid, set to fastest rate for min backpressure
        sample_rate = 192000;
        master_clock_frequency = MCLK_48;
    }
    spdif_tx_reconfigure_sample_rate(c_spdif_tx, sample_rate, master_clock_frequency);
}
#endif
